@@ -12,6 +12,7 @@
 #include "common.h"
 
 #include <cstdint>
+#include <functional>
 #include <memory>  // std::shared_ptr, std::unique_ptr
 #include <set>
 #include <string>
@@ -232,6 +233,9 @@ class SONATA_API Population
      */
     std::string _dynamicsAttributeDataType(const std::string& name) const;
 
+    template <typename T>
+    Selection filterAttribute(const std::string& name, std::function<bool(const T)> pred) const;
+
   protected:
     Population(const std::string& h5FilePath,
                const std::string& csvFilePath,
